# srvrr.tcl - SRV-aware socket command
#
# for client sockets, it supports one additional switch:
#
#       -service name
#
# e.g.,
#
#       socket -service apex-edge example.com 10290
#
# which says to use the SRV algorithm from RFC 2782. If there
# aren't any SRV RRs for the domain, then this call is identical
# to the socket command.
#


namespace eval srvrr {
    variable srvrr
    array set srvrr { enabledP 0 }
}


proc srvrr::toggle {{enableP -1}} {
    variable srvrr

    if {$enableP > 0} {
        set enableP 1
    }

    if {($enableP < 0) || ($srvrr(enabledP) == $enableP)} {
        return $srvrr(enabledP)
    }

    if {$enableP} {
        rename ::socket ::_socket
        proc ::socket {args} "eval [namespace current]::socket \$args"
    } else {
        rename ::_socket ::socket       
    }

    set srvrr(enabledP) $enableP
}


proc srvrr::socket {args} {
    variable srvrr

    if {$srvrr(enabledP)} {
        set socket [list ::_socket]
    } else {
        set socket [list ::socket]
    }

    if {[lsearch $args -server] >= 0} {
        return [eval $socket $args]
    }

    set opts {}
    set optP 0
    set asyncP 0
    set domainP 0
    set portP 0
    foreach arg $args {
        switch -- $arg {
            -myaddr
                -
            -myport
                -
            -service {
                lappend opts $arg
                set optP 1
            }

            -async {
                set asyncP 1
            }

            default {
                if {$optP} {
                    lappend opts $arg
                    set optP 0
                } elseif {!$domainP} {
                    set domain $arg
                    set domainP 1
                } elseif {!$portP} {
                    set port $arg
                    set portP 1
                } else {
                    error "wrong #args: should be socket ?args? domain port"
                }
            }
        }
    } 
    if {$optP} {
        error "no argument given for [lindex $opts end] option"
    }
    if {!$domainP} {
        error "wrong #args: should be socket ?args? domain ?port?"
    }

    if {[set x [lsearch $opts -service]] < 0} {
        return [eval $socket $args]
    }
    set service [lindex $opts [expr $x+1]]
    set opts [lreplace $opts $x [expr $x+1]]

    if {([catch { getsrv $domain $service } servers]) \
            || ([set n [llength $servers]] == 0)} {
        if {!$portP} {
            error \
                "no SRV RRs (service $service for $domain) and no default port"
        }
        if {$asyncP} {
            lappend opts -async
        }
        return [eval $socket $opts [list $domain $port]]
    }
    
    if {$n == 1} {
        array set rr [lindex $servers 0]
        if {![string compare $rr(name) ""]} {
            error "service $service disabled for $domain"
        }
    }

    foreach entry [lrange $servers 0 [expr $n-2]] {
        array set rr $entry
        if {![catch { eval $socket $opts [list $rr(name) $rr(port)] } \
                   result]} {
            return $result
        }
    }

    array set rr [lindex $servers end]
    if {$asyncP} {
        lappend opts -async
    }
    eval $socket $opts [list $rr(name) $rr(port)]
}
